import importlib
from typing import Optional, Tuple

import pytest
from packaging.version import Version


# thanks to xarray
def _importorskip(
    modname: str, minversion: Optional[str] = None
) -> Tuple[bool, pytest.MarkDecorator]:
    try:
        mod = importlib.import_module(modname)
        has = True
        if minversion is not None:  # pragma: no cover
            if Version(mod.__version__) < Version(minversion):
                raise ImportError("Minimum version not satisfied")
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


has_requests_cache, requires_requests_cache = _importorskip("requests_cache")



"""SDMX-ML 3.0.0 reader."""
from typing import Any, Dict

import sdmx.urn
from sdmx.format import Version, list_media_types
from sdmx.model import common
from sdmx.model import v30 as model

from . import v21


class Reference(v21.Reference):
    """Parse SDMX-ML 3.0 references."""

    @classmethod
    def info_from_element(cls, elem):
        try:
            result = sdmx.urn.match(elem.text)
            # If the URN doesn't specify an item ID, it is probably a reference to a
            # MaintainableArtefact, so target_id and id are the same
            result.update(target_id=result["item_id"] or result["id"])
        except ValueError:
            # Bare string that is the ID of e.g. a component (dimension)
            id = elem.text.strip()
            if id:
                result = {"id": id, "target_id": id, "class": None, "package": None}
            else:
                raise v21.NotReference

        return result


class Reader(v21.Reader):
    xml_version = Version["3.0.0"]
    media_types = list_media_types(base="xml", version=xml_version)
    Reference = Reference


# Rewrite the v21.Reader collection of parsers to refer to SDMX-ML 3.0.0 namespaces
# instead of SDMX-ML 2.1
new_parsers = dict()
for (tag, event), func in v21.Reader.parser.items():
    # Construct a new tag using the same prefix (e.g. "str") and local name
    new_tag = Reader.format.qname(
        v21.Reader.format.ns_prefix(tag.namespace), tag.localname
    )
    # Store a reference to the same function
    new_parsers[(new_tag, event)] = func
# Replace the parser collection
Reader.parser = new_parsers

# Shorthand
start = Reader.start
end = Reader.end

# In SDMX-ML 3.0, individual classes of ItemScheme are collected in separate XML
# container elements. Skip all of these.
start(
    """
    str:AgencySchemes str:ConceptSchemes str:CustomTypeSchemes str:DataConstraints
    str:GeographicCodelists str:GeoGridCodelists str:Hierarchies
    str:NamePersonalisationSchemes str:RulesetSchemes str:TransformationSchemes
    str:UserDefinedOperatorSchemes str:ValueLists str:VtlMappingSchemes
    """
)(None)

# New qnames in SDMX-ML 3.0 parsed using existing methods from .reader.xml.v21
end("str:GeoCell str:GridDefinition str:Value")(v21._text)
end("str:GeographicCodelist str:ValueList")(v21._itemscheme)
start("str:GeoFeatureSetCode str:GeoGridCode str:ValueItem", only=False)(
    v21._item_start
)
end("str:GeoFeatureSetCode str:GeoGridCode str:ValueItem", only=False)(v21._item_end)
start("str:Measure str:MetadataAttribute", only=False)(v21._component_start)
end("str:Measure str:MetadataAttribute", only=False)(v21._component_end)
end("str:MetadataAttributeList")(v21._cl)
end("str:DataConstraint")(v21._cc)
end("str:KeyValue")(v21._ms)
end("str:Observation")(v21._ar_kind)


@end("str:Codelist")
def _cl(reader, elem):
    try:
        sdmx.urn.match(elem.text)
    except ValueError:
        result = v21._itemscheme(reader, elem)
        result.extends = reader.pop_all(model.CodelistExtension)
        return result
    else:
        reader.push(elem, elem.text)


@end("str:CodelistExtension")
def _cl_ext(reader, elem):
    cs = reader.pop_all(model.CodeSelection, subclass=True) or [None]
    assert 1 == len(cs)
    return model.CodelistExtension(
        extends=reader.pop_resolved_ref("Codelist"),
        prefix=elem.attrib.get("prefix", None),
        selection=cs[0],
    )


@end("str:ExclusiveCodeSelection str:InclusiveCodeSelection")
def _code_selection(reader, elem):
    return reader.class_for_tag(elem.tag)(mv=reader.pop_all(model.MemberValue))


@end("str:MemberValue")
def _mv(reader, elem):
    return reader.model.MemberValue(value=elem.text)


@end("str:GeoGridCodelist")
def _ggcl(reader, elem):
    result = v21._itemscheme(reader, elem)
    result.grid_definition = reader.pop_single("GridDefinition")
    return result


@end("str:GeoGridCode", only=False)
def _ggc_end(reader, elem):
    result = v21._item_end(reader, elem)
    result.geo_cell = reader.pop_single("GeoCell")
    return result


# §5.3: Data Structure Definition


@end("str:AttributeRelationship")
def _ar(reader: Reader, elem):
    dsd = reader.peek("current DSD")

    refs = reader.pop_all(reader.Reference)
    if not len(refs):
        return

    # Iterate over parsed references to Components
    args: Dict[str, Any] = dict(dimensions=list())
    for ref in refs:
        # Use the <Ref id="..."> to retrieve a Component from the DSD
        if issubclass(ref.target_cls, model.DimensionComponent):
            component = dsd.dimensions.get(ref.target_id)
            args["dimensions"].append(component)
        elif ref.target_cls is model.Measure:
            # Since <str:AttributeList> occurs before <str:MeasureList>, this is
            # usually a forward reference. We *could* eventually resolve it to confirm
            # consistency (the referenced ID is same as the PrimaryMeasure.id), but
            # that doesn't affect the returned value, since PrimaryMeasureRelationship
            # has no attributes.
            return model.ObservationRelationship()
        elif ref.target_cls is model.GroupDimensionDescriptor:
            args["group_key"] = dsd.group_dimensions[ref.target_id]

    ref = reader.pop_single("AttachmentGroup")
    if ref:
        args["group_key"] = dsd.group_dimensions[ref.target_id]

    if len(args["dimensions"]):
        return common.DimensionRelationship(**args)


# §5.4: Data Set


@end(":Value")
def _complex_value(reader: Reader, elem):
    try:
        reader.push("ComplexValue", model.InternationalString(reader.pop_all("Text")))
    except Exception:  # pragma: no cover
        raise NotImplementedError


@end(":Comp")
def _complex(reader: Reader, elem):
    ds = reader.get_single("DataSet")

    assert ds is not None
    da = ds.structured_by.attributes.getdefault(elem.attrib["id"])

    reader.stack.setdefault("Attributes", {-1: {}})

    reader.stack["Attributes"][-1][da.id] = model.AttributeValue(
        value=reader.pop_all("ComplexValue"), value_for=da
    )


# §8: Hierarchy


@end("str:Hierarchy")
def _h(reader: Reader, elem):
    cls = reader.class_for_tag(elem.tag)
    return reader.maintainable(
        cls,
        elem,
        has_formal_levels=eval(elem.attrib["hasFormalLevels"].title()),
        codes={c.id: c for c in reader.pop_all(model.HierarchicalCode)},
        level=reader.pop_single(common.Level),
    )



from operator import add, sub

import sdmx
from sdmx.model.common import DEFAULT_LOCALE, Item
from sdmx.util.item_structure import parse_all, parse_item_description


def test_parse(specimen):
    with specimen("IMF/CL_AREA-structure.xml") as f:
        msg = sdmx.read_sdmx(f)

    cl = msg.codelist["CL_AREA"]

    result = parse_all(cl)

    # Result has one element per item in the scheme
    assert 901 == len(result) == len(cl)

    # 185 items have parseable expressions in their descriptions
    assert 185 == sum(1 if len(v) else 0 for v in result.values())

    # Specific expressions are available and refer to other items in the scheme
    assert [(add, cl["1A"]), (sub, cl["1C"]), (sub, cl["5B"])] == result["9B"]


def test_localizations(caplog):
    """Selection of locale of description for parsing works."""
    i = Item(
        id="FOO",
        description={DEFAULT_LOCALE: "", "zh": "FOO = B + C - D", "ru": ""},
    )

    assert 0 == len(parse_item_description(i))
    assert caplog.messages[-1].endswith("using .localized_default('en')")

    assert 3 == len(parse_item_description(i, locale="zh"))



import importlib.resources
import json
from dataclasses import dataclass, field
from enum import Enum
from importlib import import_module
from io import IOBase
from typing import Any, Dict, Optional, Tuple, Union

from requests import Response

from sdmx.model.v21 import DataStructureDefinition
from sdmx.rest import Resource

#: Data sources registered with :mod:`sdmx`.
sources: Dict[str, "Source"] = {}

#: Valid data content types for SDMX REST API messages.
DataContentType = Enum("DataContentType", "CSV JSON XML")

#: Default values for :attr:`.Source.supports`. Most of these values indicate REST API
#: endpoints that are described in the standards but are not implemented by any source
#: currently in :file:`sources.json`; these all return 404.
SDMX_ML_SUPPORTS = {
    Resource.attachementconstraint: False,
    Resource.customtypescheme: False,
    Resource.data: True,
    Resource.metadata: True,
    Resource.namepersonalisationscheme: False,
    Resource.organisationunitscheme: False,
    Resource.process: False,
    Resource.reportingtaxonomy: False,
    Resource.rulesetscheme: False,
    Resource.schema: False,
    Resource.transformationscheme: False,
    Resource.userdefinedoperatorscheme: False,
    Resource.vtlmappingscheme: False,
}


@dataclass
class Source:
    """SDMX-IM RESTDatasource.

    This class describes the location and features supported by an SDMX REST API data
    source/web service.

    It also provides three hooks, with default implementations. Subclasses may override
    the hooks in order to handle specific features of different REST web services:

    .. autosummary::
       handle_response
       finish_message
       modify_request_args
    """

    #: :attr:`~.IdentifiableArtefact.id` of the :attr:`DataProvider`.
    id: str

    #: Base URL (API endpoint) for queries.
    url: str

    #: Human-readable name of the data source.
    name: str

    #: Additional HTTP headers to supply by default with all requests.
    headers: Dict[str, Any] = field(default_factory=dict)

    #: :class:`.DataContentType` indicating the type of data returned by the source.
    data_content_type: DataContentType = DataContentType.XML

    #: Mapping from :class:`.Resource` values to :class:`bool` indicating support for
    #: SDMX-REST endpoints and features. If not supplied, the defaults from
    #: :data:`SDMX_ML_SUPPORTS` are used.
    #:
    #: Two additional keys are valid:
    #:
    #: - ``"preview"=True`` if the source supports ``?detail=serieskeysonly``.
    #:   See :meth:`.preview_data`.
    #: - ``"structure-specific data"=True`` if the source can return structure-
    #:   specific data messages.
    supports: Dict[Union[str, Resource], bool] = field(default_factory=dict)

    def __post_init__(self):
        # Sanity check: _id attribute of a subclass matches the loaded ID.
        assert getattr(self, "_id", self.id) == self.id

        # Convert an e.g. string to a DataContentType member
        if not isinstance(self.data_content_type, DataContentType):
            self.data_content_type = DataContentType[self.data_content_type]

        # Default feature support: True for sdmx_ml, False otherwise
        sdmx_ml = self.data_content_type is DataContentType.XML

        # Update mapping of supported features
        for feature in list(Resource) + ["preview", "structure-specific data"]:
            # String name of a Resource enumeration member
            f_name = getattr(feature, "value", feature)

            # In order of precedence:
            # 1. The value already in `supports`.
            # 2. A value loaded from sources.json, appearing with a string key `f_name`
            #    in `supports`.
            # 3. The value in `SDMX_ML_SUPPORTS`, if any.
            # 4. The value `sdmx_ml`.
            self.supports.setdefault(
                feature,
                self.supports.pop(f_name, SDMX_ML_SUPPORTS.get(feature, sdmx_ml)),
            )

    # Hooks
    def handle_response(
        self, response: Response, content: IOBase
    ) -> Tuple[Response, IOBase]:
        """Handle response content of unknown type.

        This hook is called by :meth:`.Client.get` *only* when the `content` cannot be
        parsed as XML or JSON.

        See :meth:`.estat.Source.handle_response` and
        :meth:`.sgr.Source.handle_response` for example implementations.
        """
        return response, content

    def finish_message(self, message, request, **kwargs):
        """Postprocess retrieved message.

        This hook is called by :meth:`.Client.get` after a :class:`.Message` object has
        been successfully parsed from the query response.

        See :meth:`.estat.Source.finish_message` for an example implementation.
        """
        return message

    def modify_request_args(self, kwargs):
        """Modify arguments used to build query URL.

        This hook is called by :meth:`.Client.get` to modify the keyword arguments
        before the query URL is built.

        The default implementation handles requests for 'structure-specific data' by
        adding an HTTP 'Accepts:' header when a 'dsd' is supplied as one of the
        `kwargs`.

        See :meth:`.sgr.Source.modify_request_args` for an example override.

        Returns
        -------
        None
        """
        if self.data_content_type is DataContentType.XML:
            dsd = kwargs.get("dsd", None)
            if isinstance(dsd, DataStructureDefinition):
                kwargs.setdefault("headers", {})
                kwargs["headers"].setdefault(
                    "Accept",
                    "application/vnd.sdmx.structurespecificdata+xml;" "version=2.1",
                )


@dataclass
class _NoSource(Source):
    pass


NoSource = _NoSource(id="", url="", name="")


def add_source(
    info: Union[Dict, str], id: Optional[str] = None, override: bool = False, **kwargs
) -> None:
    """Add a new data source.

    The *info* expected is in JSON format:

    .. code-block:: json

        {
          "id": "ESTAT",
          "documentation": "http://data.un.org/Host.aspx?Content=API",
          "url": "http://ec.europa.eu/eurostat/SDMX/diss-web/rest",
          "name": "Eurostat",
          "supports": {"codelist": false, "preview": true}
        }

    …with unspecified values using the defaults; see :class:`Source`.

    Parameters
    ----------
    info : dict-like
        String containing JSON information about a data source.
    id : str
        Identifier for the new datasource. If :obj:`None` (default), then `info['id']`
        is used.
    override : bool
        If :obj:`True`, replace any existing data source with *id*. Otherwise, raise
        :class:`ValueError`.
    **kwargs
        Optional callbacks for *handle_response* and *finish_message* hooks.

    """
    _info = json.loads(info) if isinstance(info, str) else info
    id = id or _info["id"]

    _info.update(kwargs)

    if not override and id in sources:
        raise ValueError(f"Data source {repr(id)} already defined; use override=True")

    # Maybe import a subclass that defines a hook
    SourceClass = Source
    try:
        mod = import_module("." + id.lower(), "sdmx.source")
    except ImportError:
        pass
    else:
        SourceClass = getattr(mod, "Source", None) or SourceClass

    sources[id] = SourceClass(**_info)


def list_sources():
    """Return a sorted list of valid source IDs.

    These can be used to create :class:`Client` instances.
    """
    return sorted(sources.keys())


def load_package_sources():
    """Discover all sources listed in :file:`sources.json`."""
    try:
        ref = importlib.resources.files("sdmx").joinpath("sources.json")
    except AttributeError:  # Python <3.9
        from copy import copy

        with importlib.resources.path("sdmx", "sources.json") as path:
            ref = copy(path)

    with ref.open("rb") as f:
        for info in json.load(f):
            add_source(info)


load_package_sources()



import pytest

from sdmx.format import Flag, MediaType
from sdmx.reader.base import BaseReader


class TestBaseReader:
    @pytest.fixture
    def MinimalReader(self):
        """A reader that implements the minimum abstract methods."""

        class cls(BaseReader):
            media_types = [
                MediaType("", "xml", "2.1", Flag.data, full="application/foo"),
            ]

            def read_message(self, source, dsd=None):
                pass  # pragma: no cover

        return cls

    def test_detect(self, MinimalReader):
        assert False is MinimalReader.detect(b"foo")

    def test_handles_media_type(self, caplog, MinimalReader):
        """:meth:`.handles_media_type` matches even when params differ, but logs."""
        assert True is MinimalReader.handles_media_type("application/foo; bar=qux")
        assert (
            "Match application/foo with params {'bar': 'qux'}; "
            "expected {'version': '2.1'}" in caplog.messages
        )



import logging
from dataclasses import dataclass

import pytest

import sdmx.model as model
from sdmx.model import v21
from sdmx.model.common import (
    Agency,
    AnnotableArtefact,
    Annotation,
    Contact,
    IdentifiableArtefact,
    Item,
    ItemScheme,
    MaintainableArtefact,
    NameableArtefact,
    Representation,
)


class TestAnnotation:
    def test_text(self) -> None:
        """`text` can be :class:`str`."""
        a = Annotation(text="Foo")

        assert a.text.localizations["en"] == "Foo"


class TestAnnotableArtefact:
    def test_get_annotation(self):
        aa = AnnotableArtefact(
            annotations=[
                Annotation(id="foo", text="bar"),
                Annotation(id="baz", title="baz_title", text="baz_text"),
            ]
        )

        with pytest.raises(KeyError):
            aa.get_annotation(id="bar")

        # Retrieve with 1 key
        assert "bar" == str(aa.get_annotation(id="foo").text)

        # Retrieve with 2 keys
        assert "baz_text" == str(aa.get_annotation(id="baz", title="baz_title").text)

        # Annotations are not removed
        assert 2 == len(aa.annotations)

    def test_pop_annotation(self):
        aa = AnnotableArtefact()
        anno = Annotation(id="foo", text="bar")

        assert 0 == len(aa.annotations)
        aa.annotations.append(anno)
        assert 1 == len(aa.annotations)

        with pytest.raises(KeyError):
            aa.pop_annotation(id="baz")

        assert anno == aa.pop_annotation(id="foo")
        assert 0 == len(aa.annotations)

    def test_eval_annotation(self, caplog) -> None:
        aa = AnnotableArtefact()

        value = dict(foo=1.1, bar=2, baz=True)

        aa.annotations.append(Annotation(id="test-anno0", text=repr(value)))
        aa.annotations.append(Annotation(id="test-anno1", text="value['foo']"))

        # Returns None for a missing ID
        assert None is aa.eval_annotation(id="wrong-id")

        # Entire value is retrieved
        assert value == aa.eval_annotation(id="test-anno0")

        # Cannot handle a variable reference with no globals;
        assert "value['foo']" == aa.eval_annotation(id="test-anno1")
        assert "name 'value' is not defined" in caplog.messages[0]
        caplog.clear()

        # Globals can be used if passed
        assert value["foo"] == aa.eval_annotation(
            id="test-anno1", globals=dict(value=value)
        )


class TestIdentifiableArtefact:
    def test_general(self):
        urn = (
            "urn:sdmx:org.sdmx.infomodel.conceptscheme.ConceptScheme=IT1:VARIAB_ALL"
            "(9.6)"
        )
        urn_pat = urn.replace("(", r"\(").replace(")", r"\)")

        with pytest.raises(
            ValueError, match=f"ID BAD_URN does not match URN {urn_pat}"
        ):
            IdentifiableArtefact(id="BAD_URN", urn=urn)

        # IdentifiableArtefact is hashable
        ia = IdentifiableArtefact()
        assert hash(ia) == id(ia)

        ia = IdentifiableArtefact(id="foo")
        assert hash(ia) == hash("foo")

        # Subclass is hashable
        ad = model.AttributeDescriptor()
        assert hash(ad) == id(ad)

    def test_hash_subclass(self):
        @dataclass
        class Foo(IdentifiableArtefact):
            __hash__ = IdentifiableArtefact.__hash__

        f = Foo(id="FOO")
        assert hash("FOO") == hash(f)

    def test_sort(self):
        """Test IdentifiableArtefact.__lt__."""
        # Items of the same class can be sorted
        items = [Item(id="b"), Item(id="a")]
        assert list(reversed(items)) == sorted(items)

        with pytest.raises(
            TypeError,
            match=(
                "'<' not supported between instances of 'Item' and "
                "'DataStructureDefinition'"
            ),
        ):
            sorted([v21.DataStructureDefinition(id="c")] + items)


class TestNameableArtefact:
    def test_eq(self):
        na = NameableArtefact(id="FOO", name="Foo")
        assert na == "FOO"

    def test_namea(self, caplog) -> None:
        na1 = NameableArtefact(name=dict(en="Name"), description=dict(en="Description"))
        na2 = NameableArtefact()

        assert not na1.compare(na2)
        assert caplog.messages[-1] == "Not identical: name <en: Name> != <>"

        na2.name["en"] = "Name"

        assert not na1.compare(na2)
        assert (
            caplog.messages[-1] == "Not identical: description <en: Description> != <>"
        )

        na2.description["en"] = "Description"

        assert na1.compare(na2)


class TestMaintainableArtefact:
    def test_urn(self):
        urn = (
            "urn:sdmx:org.sdmx.infomodel.conceptscheme.ConceptScheme="
            "IT1:VARIAB_ALL(9.6)"
        )
        ma = MaintainableArtefact(id="VARIAB_ALL", urn=urn)

        # Version is parsed from URN
        assert ma.version == "9.6"

        # Mismatch raises an exception
        with pytest.raises(ValueError, match="Version 9.7 does not match URN"):
            MaintainableArtefact(version="9.7", urn=urn)

        # Maintainer is parsed from URN
        assert ma.maintainer == Agency(id="IT1")

        # Mismatch raises an exception
        with pytest.raises(ValueError, match="Maintainer FOO does not match URN"):
            MaintainableArtefact(maintainer=Agency(id="FOO"), urn=urn)


class TestItem:
    def test_name(self) -> None:
        """`name` can be :class:`str`."""
        Item(name="Foo")

    def test_general(self):
        # Add a tree of 10 items
        items = []
        for i in range(10):
            items.append(Item(id="Foo {}".format(i)))

            if i > 0:
                items[-1].parent = items[-2]
                items[-2].child.append(items[-1])

        # __init__(parent=...)
        Item(id="Bar 1", parent=items[0])
        assert len(items[0].child) == 2

        # __init__(child=)
        bar2 = Item(id="Bar 2", child=[items[0]])

        # __contains__()
        assert items[0] in bar2
        assert items[-1] in items[0]

        # get_child()
        assert items[0].get_child("Foo 1") == items[1]

        with pytest.raises(ValueError):
            items[0].get_child("Foo 2")

        # Hierarchical IDs constructed automatically
        assert items[0].child[0].hierarchical_id == "Bar 2.Foo 0.Foo 1"


class TestItemScheme:
    def test_compare0(self, caplog):
        caplog.set_level(logging.DEBUG)

        is0 = ItemScheme()
        is1 = ItemScheme()

        is0.append(Item(id="foo", name="Foo"))
        is1.append(Item(id="foo", name="Bar"))

        assert not is0.compare(is1)

        # Log shows that items with same ID have different name
        assert caplog.messages[-2:] == [
            "Not identical: name <en: Foo> != <en: Bar>",
            "…for items with id='foo'",
        ]

    def test_compare1(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        is0.append(Item(id="foo"))

        is1: ItemScheme = ItemScheme(id="is0")
        is1.append(Item(id="foo"))

        assert is0.compare(is1) is True

        is2: ItemScheme = ItemScheme(id="is0")
        is2.append(Item(id="bar"))

        assert is0.compare(is2) is False

    def test_get_hierarchical(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        foo: Item = Item(id="foo")
        bar: Item = Item(id="bar")
        foo.append_child(bar)

        is0.append(foo)
        is0.append(bar)

        assert bar is is0.get_hierarchical("foo.bar")

        # With non-hierarchical ID
        assert foo is is0.get_hierarchical("foo")

        with pytest.raises(KeyError):
            is0.get_hierarchical("foo.baz")

    def test_other(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        foo0: Item = Item(id="foo0")

        # With a single Item

        # append()
        is0.append(foo0)

        # __getattr__
        assert is0.foo0 is foo0

        # __getitem__
        assert is0["foo0"] is foo0

        # __contains__
        assert "foo0" in is0
        assert foo0 in is0

        # __len__
        assert len(is0) == 1

        # __repr__
        assert repr(is0) == "<ItemScheme is0 (1 items)>"

        # __iter__
        assert all(i is foo0 for i in is0)

        # With multiple Items

        foo1: Item = Item(id="foo1")
        foo2: Item = Item(id="foo2")
        items_list = [foo0, foo1, foo2]
        items_dict = {"foo0": foo0, "foo1": foo1, "foo2": foo2}

        # set with a non-dict
        is0.items = {i.id: i for i in items_list}
        assert is0.items == items_dict

        # set with a dict
        is0.items = items_dict
        assert is0.items == items_dict

        # extend()
        is0.items = {foo0.id: foo0}
        is0.extend(items_list[1:])
        assert is0.items == items_dict

        # setdefault()
        bar0 = is0.setdefault(id="bar")
        assert bar0.id == "bar"

        with pytest.raises(ValueError):
            is0.setdefault(foo0, id="bar")

        is0.setdefault(id="bar1", parent="foo0")
        bar1 = is0.setdefault(id="bar1", parent=foo0)

        # get_hierarchical()
        assert is0.get_hierarchical("foo0.bar1") is bar1


class TestRepresentation:
    def test_repr(self) -> None:
        is0: ItemScheme = ItemScheme(id="is0")
        r = Representation(enumerated=is0)
        assert "<Representation: is0, []>" == repr(r)


class TestContact:
    def test_init(self):
        Contact(
            name="Jane Smith", org_unit="Statistics Office", responsibility="Director"
        )

        c1 = Contact()
        c2 = Contact()

        c2.name = "foo"

        assert repr(c1) != repr(c2)


class TestStructure:
    @pytest.fixture
    def obj(self):
        # Use BaseDataStructureDefinition as a concrete/less abstract subclass
        return model.BaseDataStructureDefinition()

    def test_grouping(self, obj) -> None:
        result = obj.grouping
        # Class has an AttributeDescriptor, DimensionDescriptor, and dict-like of
        # GroupDimensionDescriptor
        assert 3 == len(result)

    def test_replace_grouping(self, obj) -> None:
        class Foo(model.ComponentList):
            pass

        # Cannot replace with an instance of Foo, because this does not correspond to
        # the type of any member of the class
        with pytest.raises(TypeError, match="No grouping of type"):
            obj.replace_grouping(Foo())



import pickle
from dataclasses import dataclass
from typing import Generator, Type

import pytest

import sdmx
from sdmx.dictlike import DictLike, DictLikeDescriptor
from sdmx.util import only, parse_content_type


class TestDictLike:
    @pytest.fixture
    def Foo(self) -> Generator[Type, None, None]:
        # Example class
        @dataclass
        class Foo:
            items: DictLikeDescriptor[str, int] = DictLikeDescriptor()

        yield Foo

    def test_init(self):
        # NB use with "assert False" in DictLike.__setitem__
        DictLike(foo=1)
        DictLike((("foo", 1),))
        DictLike(dict(foo=1))

    def test_class(self):
        dl = DictLike()

        # Set by item name
        dl["TIME_PERIOD"] = 3
        dl["CURRENCY"] = "USD"

        # Access by attribute name
        assert dl.TIME_PERIOD == 3

        # Access by item index
        assert dl[1] == "USD"

        # Access beyond index
        with pytest.raises(KeyError):
            dl["FOO"]

        with pytest.raises(IndexError):
            dl[2]

        with pytest.raises(AttributeError):
            dl.FOO

        # copy() returns same class
        copied = dl.copy()
        assert isinstance(copied, DictLike)
        assert copied.TIME_PERIOD == dl.TIME_PERIOD

    def test_validation(self, Foo) -> None:
        f = Foo()
        assert type(f.items) is DictLike
        assert (str, int) == f.items._types

        # Can be set with DictLike
        f.items = DictLike(a=1, b=2)
        assert type(f.items) is DictLike

        # Can be set with dict()
        f.items = {"a": 1, "b": 2}
        assert type(f.items) is DictLike

        # Type checking on creation
        with pytest.raises(TypeError):
            f = Foo(items={1: "a"})

        # Type checking on assignment
        f = Foo()
        with pytest.raises(TypeError):
            f.items = {1: "a"}

        # Type checking on setting elements
        f = Foo(items={"a": 1})
        with pytest.raises(TypeError):
            f.items[123] = 456

        # With no initial value set
        f = Foo()
        with pytest.raises(TypeError):
            f.items[123] = 456

    def test_compare(self, caplog):
        dl1 = DictLike(a="foo", b="bar")
        dl2 = DictLike(c="baz", a="foo")

        assert not dl1.compare(dl2)
        assert "Not identical: ['a', 'b'] / ['a', 'c']" in caplog.messages

    def test_pickle(self, specimen):
        """Instances included in a Pydantic model can be pickled."""
        with specimen("sg-xs.xml") as f:
            msg1 = sdmx.read_sdmx(f)

        value = pickle.dumps(msg1)
        msg2 = pickle.loads(value)

        assert msg1.compare(msg2)


def test_only():
    assert None is only(filter(lambda x: x == "foo", ["bar", "baz"]))
    assert None is only(filter(lambda x: x == "foo", ["foo", "bar", "foo"]))


def test_parse_content_type():
    """:func:`.parse_content_type` handles whitespace, quoting, and empty params."""
    assert (
        "application/foo",
        dict(version="1.2", charset="UTF-8"),
    ) == parse_content_type("application/foo; version = 1.2 ; ;charset='UTF-8'")



"""Information about the SDMX-ML 2.1 file format."""
from sdmx.model import v21

from .common import XMLFormat

FORMAT = XMLFormat(
    model=v21,
    base_ns="http://www.sdmx.org/resources/sdmxml/schemas/v2_1",
    class_tag=[
        ("message.DataMessage", "mes:GenericData"),
        ("message.DataMessage", "mes:GenericTimeSeriesData"),
        ("message.DataMessage", "mes:StructureSpecificTimeSeriesData"),
        ("model.NoSpecifiedRelationship", "str:None"),
        ("model.DataflowDefinition", "str:Dataflow"),
        ("model.DimensionDescriptorValuesTarget", "str:KeyDescriptorValuesTarget"),
        ("model.MetadataflowDefinition", "str:Metadataflow"),
        ("model.MetadataSet", "mes:MetadataSet"),
        ("model.ReportedAttribute", "md:ReportedAttribute"),
        ("model.TargetIdentifiableObject", ":ObjectReference"),
        ("model.TargetIdentifiableObject", "md:ObjectReference"),
        ("model.TargetObjectKey", ":Target"),
        ("model.TargetObjectKey", "md:Target"),
        ("model.TargetReportPeriod", "ReportPeriod"),
        ("model.TargetReportPeriod", "md:ReportPeriod"),
        ("model.MetadataReport", ":Report"),
        ("model.MetadataReport", "md:Report"),
    ]
    + [
        (f"model.{name}", f"str:{name}")
        for name in """
            CodelistMap
            CodeMap
            ContentConstraint
            HierarchicalCodelist
            Hierarchy
            IdentifiableObjectTarget
            MeasureDimension
            MetadataAttribute
            MetadataTarget
            PrimaryMeasure
            ReportPeriodTarget
            ReportStructure
            StructureSet
        """.split()
    ],
)


def __getattr__(name):
    return getattr(FORMAT, name)



import re

from . import Source as BaseSource

re_500 = re.compile(r"(An error has occurred)\.")


class Source(BaseSource):
    _id = "ABS"

    def handle_response(self, response, content):
        """Handle ABS' own text/html error page for some endpoints."""
        ctype = response.headers.get("content-type", "")
        if "text/html" in ctype:
            buf = ""
            while True:
                chunk = content.read().decode()
                if len(chunk) == 0:
                    break

                buf += chunk

                match = re_500.search(buf)
                if match:
                    # Overwrite the original response
                    (response.reason,) = match.groups()
                    response.status_code = 500
                    response.raise_for_status()

        return super().handle_response(response, content)



# Configuration file for the Sphinx documentation builder.
#
# For a full list of options, see
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information ---------------------------------------------------------------

project = "sdmx"
copyright = "2014–2024 sdmx1 developers"


# -- General configuration -------------------------------------------------------------

# Sphinx extension module names
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

nitpicky = True

rst_prolog = """
.. role:: py(code)
   :language: python
.. role:: xml(code)
   :language: xml
"""

# -- Options for HTML output -----------------------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = "sphinx_book_theme"

html_theme_options = dict(
    path_to_docs="doc",
    repository_url="https://github.com/khaeru/sdmx",
    show_navbar_depth=2,
    use_edit_page_button=True,
    use_issues_button=True,
    use_repository_button=True,
    use_source_button=True,
)


# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/khaeru/sdmx/issues/%s", "#%s"),
    "pull": ("https://github.com/khaeru/sdmx/pull/%s", "PR #%s"),
    "gh-user": ("https://github.com/%s", "@%s"),
}


# -- Options for sphinx.ext.intersphinx ------------------------------------------------

intersphinx_mapping = {
    "np": ("https://numpy.org/doc/stable/", None),
    "pd": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "py": ("https://docs.python.org/3/", None),
    "requests": ("https://requests.readthedocs.io/en/latest/", None),
    "requests-cache": ("https://requests-cache.readthedocs.io/en/latest/", None),
}

# -- Options for sphinx.ext.linkcode ---------------------------------------------------


def linkcode_resolve(domain, info):
    if domain != "py" or not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return f"https://github.com/khaeru/sdmx/tree/main/{filename}.py"


# -- Options for sphinx.ext.napoleon ---------------------------------------------------

napolean_preprocess_types = True

# -- Options for sphinx.ext.todo -------------------------------------------------------

# If True, todo and todolist produce output, else they produce nothing
todo_include_todos = True

# -- Options for IPython.sphinxext.ipython_directive -----------------------------------

# Specify if the embedded Sphinx shell should import Matplotlib and set the backend
ipython_mplbackend = ""

# -- Dynamic configuration


def setup(app):
    from importlib import import_module
    from pathlib import Path

    # Update files containing lists of classes
    for mod in "common", "v21", "v30":
        # prefix = f"{mod}." if mod != "common" else ""
        prefix = f"{mod}."
        path = Path(__file__).parent.joinpath("api", f"model-{mod}-list.rst")
        with open(path, "w") as f:
            f.write(".. This file is auto-generated by doc/conf.py.\n\n")
            f.writelines(
                f":obj:`~.{prefix}{name}`\n"
                for name in sorted(import_module(f"sdmx.model.{mod}").__all__)
            )



"""Tests against the actual APIs for specific data sources.

HTTP responses from the data sources are cached in tests/data/cache.
To force the data to be retrieved over the Internet, delete this directory.
"""
# TODO add a pytest argument for clearing this cache in conftest.py
import logging
from pathlib import Path
from typing import Any, Dict, Tuple, Type, Union

import pytest
import requests_mock

import sdmx
from sdmx import Client
from sdmx.exceptions import HTTPError, XMLParseError

# Mark the whole file so the tests can be excluded/included
pytestmark = pytest.mark.source

log = logging.getLogger(__name__)


NI = "Not implemented in sdmx1"


class DataSourceTest:
    """Base class for data source tests."""

    # TODO also test structure-specific data

    #: Must be one of the IDs in sources.json.
    source_id: str

    #: Failures affecting **all** data sources, internal to :mod:`sdmx`.
    xfail_common: Dict[str, Any] = {}

    #: Mapping of endpoint → Exception subclass. Tests of these endpoints are expected
    #: to fail with the given kind of exception.
    xfail: Dict[str, Union[Type[Exception], Tuple[Type[Exception], str]]] = {}

    #: True to xfail if a 503 Error is returned.
    tolerate_503 = False

    #: Keyword arguments for particular endpoints.
    endpoint_args: Dict[str, Dict[str, Any]] = {}

    @pytest.fixture
    def cache_path(self, test_data_path):
        # Use a common cache file for all agency tests
        (test_data_path / ".cache").mkdir(exist_ok=True)

        yield test_data_path / ".cache" / self.source_id

    @pytest.fixture
    def client(self, cache_path):
        return Client(self.source_id, cache_name=str(cache_path), backend="sqlite")

    # NB the following can be added to any subclass below for SSL failures. Update the
    #    docstring to describe the nature of the problem.
    # @pytest.fixture
    # def client(self, cache_path):
    #     """Identical to DataSourceTest, except skip SSL certificate verification.
    #
    #     As of [DATE], this source returns an invalid certificate.
    #     """
    #     return Client(
    #         self.source_id, cache_name=str(cache_path), backend="sqlite", verify=False
    #     )

    @pytest.mark.network
    def test_endpoint(self, pytestconfig, cache_path, client, endpoint, args):
        # See sdmx.testing._generate_endpoint_tests() for values of `endpoint`
        cache = cache_path.with_suffix(f".{endpoint}.xml")

        try:
            message = client.get(endpoint, tofile=cache, **args)
        except HTTPError as e:  # For debugging/test development
            print(e)
            raise

        if pytestconfig.getoption("verbose"):
            print(repr(message))
            if endpoint == "dataflow":
                # Display the IDs of data flows; this can be used to identify targets
                # for tests of the data endpoint for a particular source
                for dfd in message.dataflow:
                    print(repr(dfd))

        # For debugging
        # print(cache, cache.read_text(), result, sep="\n\n")
        # assert False

        # All parsed contents can also be converted to pandas
        sdmx.to_pandas(message)


class TestABS(DataSourceTest):
    source_id = "ABS"

    endpoint_args = {
        "data": dict(
            resource_id="ABS,ANA_AGG,1.0.0",
            key="....Q",
            params=dict(startPeriod="2020-Q1", endPeriod="2022-Q4"),
        )
    }


class TestABS_JSON(DataSourceTest):
    source_id = "ABS_JSON"

    endpoint_args = {
        "data": dict(
            resource_id="ABS,ANA_AGG,1.0.0",
            key="....Q",
        )
    }


class TestBBK(DataSourceTest):
    source_id = "BBK"

    # See https://github.com/khaeru/sdmx/issues/82
    xfail = {
        "codelist": XMLParseError,
        "datastructure": XMLParseError,
    }

    endpoint_args = {
        "data": dict(
            resource_id="BBDB2",
            key="H.DE.Y.A.C.IFRS.B.A.K.E.E001.VGH.A",
            params=dict(startPeriod="2017-S1", endPeriod="2019-S1"),
        ),
        "codelist": dict(
            resource_id="CL_BBK_ERX_RATE_TYPE", params=dict(references="none")
        ),
        "conceptscheme": dict(
            resource_id="CS_BBK_ERX",
            params=dict(detail="allstubs", references="parentsandsiblings"),
        ),
        "dataflow": dict(resource_id="BBEX3", params=dict(detail="referencestubs")),
        "datastructure": dict(resource_id="BBK_QFS"),
    }


class TestBIS(DataSourceTest):
    source_id = "BIS"

    endpoint_args = {
        "actualconstraint": dict(resource_id="CBP_D_24D"),
    }


class TestECB(DataSourceTest):
    source_id = "ECB"


# Data for requests_mock; see TestESTAT.mock()
estat_mock = {
    "http://ec.europa.eu/eurostat/SDMX/diss-web/rest/data/nama_10_gdp/..B1GQ+P3.": {
        "body": Path("ESTAT", "footer2.xml"),
        "headers": {
            "Content-Type": "application/vnd.sdmx.genericdata+xml; version=2.1"
        },
    },
    "http://ec.europa.eu/eurostat/SDMX/diss-web/file/7JUdWyAy4fmjBSWT": {
        # This file is a trimmed version of the actual response for the above query
        "body": Path("ESTAT", "footer2.zip"),
        "headers": {"Content-Type": "application/octet-stream"},
    },
}


class TestESTAT(DataSourceTest):
    source_id = "ESTAT"

    @pytest.fixture
    def mock(self, test_data_path):
        # Prepare the mock requests
        fixture = requests_mock.Mocker()
        for url, args in estat_mock.items():
            # str() here is for Python 3.5 compatibility
            args["body"] = open(str(test_data_path / args["body"]), "rb")
            fixture.get(url, **args)

        return fixture

    @pytest.mark.network
    def test_xml_footer(self, mock):
        client = Client(self.source_id)

        with mock:
            msg = client.get(url=list(estat_mock.keys())[0], get_footer_url=(1, 1))

        assert len(msg.data[0].obs) == 43

    @pytest.mark.network
    def test_ss_data(self, client):
        """Test a request for structure-specific data.

        Examples from:
        https://ec.europa.eu/eurostat/web/sdmx-web-services/example-queries
        """
        df_id = "NAMA_10_GDP"
        args = dict(resource_id=df_id)

        # Query for the DSD
        dsd = client.dataflow(**args).dataflow[df_id].structure

        # Even with ?references=all, ESTAT returns a short message with the DSD as an
        # external reference. Query again to get its actual contents.
        if dsd.is_external_reference:
            dsd = client.get(resource=dsd).structure[0]
            log.info(repr(dsd))

        assert not dsd.is_external_reference

        # Example query, using the DSD already retrieved
        args.update(
            dict(
                key=dict(unit=["CP_MEUR"], na_item=["B1GQ"], geo=["LU"]),
                params={"startPeriod": "2012", "endPeriod": "2015"},
                dsd=dsd,
                # commented: for debugging
                # tofile="temp.xml",
            )
        )
        client.data(**args)

    @pytest.mark.network
    def test_gh_116(self, caplog, cache_path, client):
        """Test of https://github.com/khaeru/sdmx/issues/116.

        See also
        --------
        .test_reader_xml.test_gh_116
        """
        msg = client.get(
            "dataflow", "GOV_10Q_GGNFA", params=dict(detail="referencepartial")
        )

        # Both versions of the GEO codelist are accessible in the message
        cl1 = msg.codelist["ESTAT:GEO(13.0)"]
        cl2 = msg.codelist["ESTAT:GEO(13.1)"]

        # cl1 is complete and items are available
        assert not cl1.is_partial and 0 < len(cl1)
        # cl2 is partial, and fewer codes are included than in cl1
        assert cl2.is_partial and 0 < len(cl2) < len(cl1)

        cl3 = msg.codelist["ESTAT:UNIT(15.1)"]
        cl4 = msg.codelist["ESTAT:UNIT(15.2)"]

        # cl3 is complete and items are available
        assert not cl3.is_partial and 0 < len(cl3)
        # cl4 is partial, and fewer codes are included than in cl1
        assert cl4.is_partial and 0 < len(cl4) < len(cl3)


class TestESTAT_COMEXT(DataSourceTest):
    source_id = "ESTAT_COMEXT"

    endpoint_args = {
        "data": dict(
            resource_id="DS-059271",
            params=dict(startPeriod="2023"),
        ),
    }


class TestCOMP(DataSourceTest):
    source_id = "COMP"

    endpoint_args = {
        "data": dict(resource_id="AID_RAIL"),
    }


class TestEMPL(DataSourceTest):
    source_id = "EMPL"

    endpoint_args = {
        "data": dict(resource_id="LMP_IND_EXP"),
    }


class TestGROW(DataSourceTest):
    source_id = "GROW"

    endpoint_args = {
        "data": dict(resource_id="POST_CUBE1_X"),
    }


class TestILO(DataSourceTest):
    source_id = "ILO"
    xfail = {
        "agencyscheme": HTTPError,  # 400
        # TODO provide endpoint_args for the following 3 to select 1 or a few objects
        "codelist": HTTPError,  # 413 Client Error: Payload Too Large
        "contentconstraint": HTTPError,  # 413 Client Error: Payload Too Large
        "datastructure": HTTPError,  # 413 Client Error: Payload Too Large
        "organisationscheme": HTTPError,  # 400
        "structure": HTTPError,  # 400
        "structureset": NotImplementedError,  # 501
    }

    @pytest.mark.network
    def test_codelist(self, cache_path, client):
        client.get(
            "codelist", "CL_ECO", tofile=cache_path.with_suffix("." + "codelist-CL_ECO")
        )

    @pytest.mark.network
    def test_gh_96(self, caplog, cache_path, client):
        client.get("codelist", "CL_ECO", params=dict(references="parentsandsiblings"))
        assert (
            "ILO does not support references='parentsandsiblings'; discarded"
            in caplog.messages
        )


class TestIMF(DataSourceTest):
    source_id = "IMF"


class TestINEGI(DataSourceTest):
    source_id = "INEGI"

    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
        "structureset": NotImplementedError,  # 501
    }

    endpoint_args = dict(
        # 404 Not Found when the own source's ID ("INEGI") is used
        conceptscheme=dict(provider="ALL")
    )


class TestINSEE(DataSourceTest):
    source_id = "INSEE"

    xfail = {
        "agencyscheme": HTTPError,  # 400
        "contentconstraint": HTTPError,  # 400
        "organisationscheme": HTTPError,  # 400
        "structure": HTTPError,  # 400
        "structureset": HTTPError,  # 400
    }


class TestISTAT(DataSourceTest):
    source_id = "ISTAT"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }
    endpoint_args = {
        "actualconstraint": dict(resource_id="CONS_92_143"),
    }

    @pytest.mark.network
    def test_gh_75(self, client):
        """Test of https://github.com/dr-leo/pandaSDMX/pull/75.

        - As of the original report on 2019-06-02, the 4th dimension was ``TIPO_DATO``,
          and the 5th ``TIPO_GESTIONE``.
        - As of 2021-01-30, these are transposed, and the 4th dimension name is
          ``TIPO_GEST``.
        - As of 2022-11-03, the dataflow uses yet another new DSD with the ID
          IT1:DCIS_SERVSOCEDU1(1.0).
        """

        df_id = "47_850"

        # Reported Dataflow query works
        # Without references="datastructure", this is a very slow query
        df = client.dataflow(df_id, params={"references": "datastructure"}).dataflow[
            df_id
        ]

        # dict() key for the query
        data_key = dict(
            FREQ=["A"],
            REF_AREA=["001001+001002"],
            DATA_TYPE=["AUTP"],
            TYPE_SOCIO_EDUC_SERVICE=["ALL"],
            TYPE_MANAGEMENT=["ALL"],
            HOLDER_SECTOR_PUBPRIV=["1"],
        )

        # Dimension components are in the correct order
        assert [dim.id for dim in df.structure.dimensions.components] == list(
            data_key.keys()
        ) + ["TIME_PERIOD"]

        # Reported data query works
        # NB the reported query key was "A.001001+001002.1.AUTP.ALL.ALL"; adjusted per
        #    the DSD change (above).
        client.data(df_id, key="A.001001+001002.AUTP.ALL.ALL.1")

        # Use a dict() key to force Client to make a sub-query for the DSD
        client.data(df_id, key=data_key)

    @pytest.mark.network
    def test_gh_104(self, client):
        """Test of https://github.com/khaeru/sdmx/issues/104.

        See also
        --------
        .test_reader_xml.test_gh_104
        """
        df_id = "22_289"

        dsd = (
            client.dataflow(df_id, params={"references": "datastructure"})
            .dataflow[df_id]
            .structure
        )

        # Data message is successfully parsed
        message = client.data(
            df_id,
            key=dict(AGE="TOTAL", SEX=["1", "2"], MARITAL_STATUS="99", REF_AREA="IT"),
            dsd=dsd,
            # tofile="debug.xml",
        )
        # Provided DSD is used to structure the data set(s) in the message
        assert dsd is message.data[0].structured_by


class TestLSD(DataSourceTest):
    source_id = "LSD"
    endpoint_args = {
        # Using the example from the documentation
        "data": dict(
            resource_id="S3R629_M3010217",
            params=dict(startPeriod="2005-01", endPeriod="2007-01"),
        )
    }

    @pytest.fixture
    def client(self, cache_path):
        """Identical to DataSourceTest, except skip SSL certificate verification.

        As of 2021-01-30, this source returns a certificate that is treated as invalid
        by the GitHub Actions job runner; but *not* on a local machine.
        """
        return Client(
            self.source_id, cache_name=str(cache_path), backend="sqlite", verify=False
        )


class TestNB(DataSourceTest):
    """Norges Bank.

    This source returns a valid SDMX Error message (100 No Results Found) for the
    'categoryscheme' endpoint.
    """

    source_id = "NB"


class TestNBB(DataSourceTest):
    source_id = "NBB"
    endpoint_args = {
        "data": dict(
            resource_id="REGPOP",
            key="POPULA.000.",
            params=dict(startTime=2013, endTime=2017),
        )
    }


class TestOECD(DataSourceTest):
    source_id = "OECD"
    endpoint_args = {
        "actualconstraint": dict(resource_id="CR_A_DSD_DEBT_TRANS_COLL@DF_MICRO"),
        "data": dict(
            resource_id="DSD_MSTI@DF_MSTI",
            headers={"Accept-Encoding": "compress, gzip"},
        ),
    }


class TestOECD_JSON(DataSourceTest):
    source_id = "OECD_JSON"

    endpoint_args = {
        "data": dict(
            resource_id="ITF_GOODS_TRANSPORT",
            key=".T-CONT-RL-TEU+T-CONT-RL-TON",
            # Uncomment this to test https://github.com/khaeru/sdmx/issues/64; also
            # covered by a specimen in sdmx-test-data
            # resource_id="PART2",
        )
    }

    @pytest.fixture
    def client(self, cache_path):
        # Same as the default implementation, only using oecd_json.Client
        from sdmx.source.oecd_json import Client

        return Client(self.source_id, cache_name=str(cache_path), backend="sqlite")


class TestSGR(DataSourceTest):
    source_id = "SGR"


class TestSPC(DataSourceTest):
    source_id = "SPC"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }
    endpoint_args = {
        "actualconstraint": dict(resource_id="CR_A_DF_ADBKI"),
        "data": dict(
            resource_id="DF_CPI",
            key="A.CK+FJ..",
            params=dict(startPeriod=2010, endPeriod=2015),
        ),
    }


class TestSTAT_EE(DataSourceTest):
    source_id = "STAT_EE"
    endpoint_args = {
        # Using the example from the documentation
        "data": dict(
            resource_id="VK12",
            key="TRD_VAL+TRD_VAL_PREV..TOTAL.A",
            params=dict(startTime=2013, endTime=2017),
        )
    }


class TestUNESCO(DataSourceTest):
    """UNESCO.

    Most endpoints are marked XFAIL because the service requires registration.
    """

    source_id = "UNESCO"
    xfail = {
        # Requires registration
        "categoryscheme": HTTPError,
        "codelist": HTTPError,
        "conceptscheme": HTTPError,
        "dataflow": HTTPError,
        "provisionagreement": HTTPError,
        # Because 'supports_series_keys_only' was set
        # TODO check
        # 'datastructure': NotImplementedError,
    }


class TestUNICEF(DataSourceTest):
    source_id = "UNICEF"

    @pytest.mark.network
    def test_data(self, client):
        dsd = client.dataflow("GLOBAL_DATAFLOW").structure[0]
        client.data("GLOBAL_DATAFLOW", key="ALB+DZA.MNCH_INSTDEL.", dsd=dsd)

    @pytest.mark.network
    def test_cd2030(self, client):
        """Test that :ref:`Countdown to 2030 <CD2030>` data can be queried."""
        dsd = client.dataflow("CONSOLIDATED", provider="CD2030").structure[0]

        # D5: Births
        client.data("CONSOLIDATED", key=dict(INDICATOR="D5"), dsd=dsd)


class TestUNSD(DataSourceTest):
    source_id = "UNSD"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }


class TestWB(DataSourceTest):
    source_id = "WB"
    xfail = {
        "contentconstraint": NotImplementedError,  # 501
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
        "structureset": NotImplementedError,  # 501
    }

    @pytest.mark.network
    def test_gh_78(self, client):
        """Test of https://github.com/khaeru/sdmx/78.

        This response required adding support for ``<mes:Department>`` and
        ``<mes:Role>`` to :mod:`.reader.xml`.
        """
        client.data("DF_WITS_Tariff_TRAINS", key=".840.000.020110.reported")


class TestWB_WDI(DataSourceTest):
    source_id = "WB_WDI"

    endpoint_args = {
        # Example from the documentation website
        "data": dict(
            resource_id="WDI",
            key="A.SP_POP_TOTL.AFG",
            params=dict(startPeriod="2011", endPeriod="2011"),
        )
    }



from warnings import warn

from . import common, v21

WARNED = set()


def __dir__():
    return dir(common)


def __getattr__(name):
    # Common classes are available directly
    try:
        return getattr(common, name)
    except AttributeError:
        pass

    try:
        result = getattr(v21, name)
    except AttributeError:
        raise
    else:
        if name not in WARNED:
            WARNED.add(name)
            warn(
                message=f"Importing {name} from sdmx.model. "
                f'Use "from sdmx.model.v21 import {name}" instead.',
                category=DeprecationWarning,
                stacklevel=2,
            )
        return result



from io import BufferedIOBase, BytesIO
from operator import itemgetter

from sdmx.util import HAS_REQUESTS_CACHE, MaybeCachedSession

#: Known keyword arguments for requests_cache.CachedSession.
CACHE_KW = [
    "allowable_codes",
    "allowable_methods",
    "backend",
    "cache_name",
    "expire_after",
    "extension",
    "fast_save",
    "location",
    "old_data_on_error",
]


class Session(metaclass=MaybeCachedSession):
    """:class:`requests.Session` subclass with optional caching.

    If :mod:`requests_cache` is installed, this class inherits from
    :class:`~.requests_cache.CachedSession` and caches responses.

    Parameters
    ----------
    timeout : float
        Timeout in seconds, used for every request.

    Other parameters
    ----------------
    kwargs :
        Values for any attributes of :class:`requests.Session`, e.g.
        :attr:`~requests.Session.proxies`,
        :attr:`~requests.Session.stream`, or
        :attr:`~requests.Session.verify`.

    Raises
    ------
    TypeError
        if :mod:`requests_cache` is *not* installed and any parameters are passed.

    """

    def __init__(self, timeout=30.1, **kwargs):
        # Separate keyword arguments for CachedSession
        cache_kwargs = dict(
            filter(itemgetter(1), [(k, kwargs.pop(k, None)) for k in CACHE_KW])
        )

        if HAS_REQUESTS_CACHE:
            # Using requests_cache.CachedSession

            # No cache keyword arguments supplied = don't use the cache
            disabled = not len(cache_kwargs.keys())

            if disabled:
                # Avoid creating any file
                cache_kwargs.setdefault("backend", "memory")

            super(Session, self).__init__(**cache_kwargs)

            # Overwrite value from requests_cache.CachedSession.__init__()
            self._is_cache_disabled = disabled
        elif len(cache_kwargs):  # pragma: no cover
            raise TypeError(
                "Arguments not supported without requests_session installed: "
                + repr(cache_kwargs)
            )
        else:  # pragma: no cover
            # Plain requests.Session: no arguments
            super(Session, self).__init__()

        # Store timeout; not a property of requests.Session
        self.timeout = timeout

        # Addition keyword arguments must match existing attributes of requests.Session
        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)


class ResponseIO(BufferedIOBase):
    """Buffered wrapper for :class:`requests.Response` with optional file output.

    :class:`ResponseIO` wraps a :class:`requests.Response` object's 'content'
    attribute, providing a file-like object from which bytes can be :meth:`read`
    incrementally.

    Parameters
    ----------
    response : :class:`requests.Response`
        HTTP response to wrap.
    tee : binary, writable :py:class:`io.BufferedIOBase`, defaults to io.BytesIO()
        *tee* is exposed as *self.tee* and not closed explicitly.
    """

    def __init__(self, response, tee=None):
        self.response = response
        if tee is None:
            tee = BytesIO()
        # If tee is a file-like object or tempfile, then use it as cache
        if isinstance(tee, BufferedIOBase) or hasattr(tee, "file"):
            self.tee = tee
        else:
            # So tee must be str or os.FilePath
            self.tee = open(tee, "w+b")
        self.tee.write(response.content)
        self.tee.seek(0)

    def readable(self):
        return True

    def read(self, size=-1):
        """Read and return up to `size` bytes by calling ``self.tee.read()``."""
        return self.tee.read(size)



import logging
from tempfile import NamedTemporaryFile
from time import sleep
from urllib.parse import urlparse
from zipfile import ZipFile

import requests

from sdmx.rest import Resource
from sdmx.source import Source as BaseSource

log = logging.getLogger(__name__)


class Source(BaseSource):
    """Handle Eurostat's mechanism for large datasets and other quirks.

    For some requests, ESTAT returns a DataMessage that has no content except for a
    ``<footer:Footer>`` element containing a URL where the data will be made available
    as a ZIP file.

    To configure :meth:`finish_message`, pass its `get_footer_url` argument to
    :meth:`.Client.get`.

    .. versionadded:: 0.2.1

    See also
    --------
    :meth:`modify_request_args`
    """

    _id = "ESTAT"

    def modify_request_args(self, kwargs):
        """Modify arguments used to build query URL.

        For the "references" query parameter, ESTAT (as of 2022-11-13) only supports the
        values "children", "descendants", or "none". Other values—including the "all" or
        "parentsandsiblings" used as defaults by :class:`.Client` cause errors. Replace
        unsupported values with "none", and use "descendants" as default.

        See also
        --------
        :pull:`107`, :pull:`108`
        """
        super().modify_request_args(kwargs)

        kwargs.pop("get_footer_url", None)

        resource_type = kwargs.get("resource_type")

        # Handle the ?references= query parameter
        params = kwargs.setdefault("params", {})
        references = params.get("references")
        if references is None:
            # Client._request_from_args() sets "all" or "parentsandsiblings" by default.
            # Neither of these values is supported by ESTAT; use "descendants" instead.
            if (
                resource_type
                in (Resource.categoryscheme, Resource.dataflow, Resource.datastructure)
                and kwargs.get("resource_id")
            ) or kwargs.get("resource"):
                params["references"] = "descendants"
        elif references not in ("children", "descendants", "none"):
            log.info(f"Replace unsupported references={references!r} with 'none'")
            params["references"] = "none"

    def finish_message(self, message, request, get_footer_url=(30, 3), **kwargs):
        """Handle the initial response.

        This hook identifies the URL in the footer of the initial response,
        makes a second request (polling as indicated by *get_footer_url*), and
        returns a new DataMessage with the parsed content.

        Parameters
        ----------
        get_footer_url : (int, int)
            Tuple of the form (`seconds`, `attempts`), controlling the interval
            between attempts to retrieve the data from the URL, and the
            maximum number of attempts to make.
        """
        # Check the message footer for a text element that is a valid URL
        url = None
        for text in getattr(message.footer, "text", []):
            if urlparse(str(text)).scheme:
                url = str(text)
                break

        if not url:
            return message

        # Unpack arguments
        wait_seconds, attempts = get_footer_url

        # Create a temporary file to store the ZIP response
        ntf = NamedTemporaryFile(prefix="pandasdmx-")
        # Make a limited number of attempts to retrieve the file
        for a in range(attempts):
            sleep(wait_seconds)
            try:
                # This line succeeds if the file exists; the ZIP response
                # is stored to ntf, and then used by the
                # handle_response() hook below
                return request.get(url=url, tofile=ntf)
            except requests.HTTPError:
                raise
        ntf.close()
        raise RuntimeError("Maximum attempts exceeded")

    def handle_response(self, response, content):
        """Handle the polled response.

        The request for the indicated ZIP file URL returns an octet-stream;
        this handler saves it, opens it, and returns the content of the single
        contained XML file.
        """

        if response.headers["content-type"] != "application/octet-stream":
            return response, content

        # Read all the input, forcing it to be copied to
        # content.tee_filename
        while True:
            if len(content.read()) == 0:
                break

        # Open the zip archive
        with ZipFile(content.tee, mode="r") as zf:
            # The archive should contain only one file
            infolist = zf.infolist()
            assert len(infolist) == 1

            # Set the new content type
            response.headers["content-type"] = "application/xml"

            # Use the unzipped archive member as the response content
            return response, zf.open(infolist[0])



import logging

import sdmx.format.protobuf_pb2 as pb

log = logging.getLogger(__name__)


def write(obj, *args, **kwargs):
    """Convert an SDMX *obj* to protobuf string."""
    return _write(obj, *args, **kwargs).SerializeToString()


def _write(obj, *args, **kwargs):
    """Helper for :meth:`write`; returns :mod:`protobuf` object(s)."""
    cls_name = obj.__class__.__name__
    func_name = f"write_{cls_name.lower()}"
    try:
        func = globals()[func_name]
    except KeyError:
        raise NotImplementedError(f"write {cls_name} to protobuf")
    else:
        return func(obj, *args, **kwargs)


def _copy(obj, pb_obj):
    """Update the attributes of *pb_obj* from the sdmx.message/.model *obj*."""
    dir_logged = False

    for attr, value in obj.__dict__.items():
        if not value:
            continue

        try:
            setattr(pb_obj, attr, value)
            log.info(f"Set {attr}")
        except Exception as exc:
            log.error(f"Failed to set {attr}: {exc}")

            if not dir_logged:
                fields = filter(lambda n: not n.startswith("_"), dir(pb_obj))
                log.info(sorted(fields))
                dir_logged = True


def write_structuremessage(obj, *args, **kwargs):
    envelope = pb.Envelope()

    for cl in obj.codelist.values():
        pb_obj = envelope.data.codelists.add()
        _copy(cl, pb_obj)

    return envelope



import re
from typing import Dict

from sdmx.model import PACKAGE, MaintainableArtefact

#: Regular expression for URNs.
URN = re.compile(
    r"urn:sdmx:org\.sdmx\.infomodel"
    r"\.(?P<package>[^\.]*)"
    r"\.(?P<class>[^=]*)=((?P<agency>[^:]*):)?"
    r"(?P<id>[^\(]*)(\((?P<version>[\d\.]*)\))?"
    r"(\.(?P<item_id>.*))?"
)

_BASE = (
    "urn:sdmx:org.sdmx.infomodel.{package}.{obj.__class__.__name__}="
    "{ma.maintainer.id}:{ma.id}({ma.version}){extra_id}"
)


def make(obj, maintainable_parent=None, strict=False):
    """Create an SDMX URN for `obj`.

    If `obj` is not :class:`.MaintainableArtefact`, then `maintainable_parent`
    must be supplied in order to construct the URN.
    """
    if not isinstance(obj, MaintainableArtefact):
        ma = maintainable_parent or obj.get_scheme()
        extra_id = f".{obj.id}"
    else:
        ma = obj
        extra_id = ""

    if not isinstance(ma, MaintainableArtefact):
        raise ValueError(
            f"Neither {repr(obj)} nor {repr(maintainable_parent)} are maintainable"
        )
    elif ma.maintainer is None:
        raise ValueError(f"Cannot construct URN for {repr(ma)} without maintainer")
    elif strict and ma.version is None:
        raise ValueError(f"Cannot construct URN for {repr(ma)} without version")

    return _BASE.format(
        package=PACKAGE[obj.__class__.__name__], obj=obj, ma=ma, extra_id=extra_id
    )


def match(value: str) -> Dict[str, str]:
    """Match :data:`URN` in `value`, returning a :class:`dict` with the match groups."""
    try:
        match = URN.match(value)
        assert match is not None
    except AssertionError:
        raise ValueError(f"not a valid SDMX URN: {value}")
    else:
        return match.groupdict()



from pathlib import Path

from . import json, xml

#: Reader classes
READERS = [json.Reader, xml.Reader]


def _readers():
    return ", ".join(map(lambda cls: cls.__name__, READERS))


def detect_content_reader(content):
    """Return a reader class for `content`.

    The :meth:`.BaseReader.detect` method for each class in :data:`READERS` is called;
    if a reader signals that it is compatible with `content`, then that class is
    returned.

    Raises
    ------
    ValueError
        If no reader class matches.
    """
    for cls in READERS:
        if cls.detect(content):
            return cls

    raise ValueError(f"{repr(content)} not recognized by any of {_readers()}")


def get_reader_for_media_type(value):
    """Return a reader class for HTTP content/media type `value`.

    Raises
    ------
    ValueError
        If no reader class matches.

    See also
    --------
    BaseReader.media_type
    """
    for cls in READERS:
        if cls.handles_media_type(value):
            return cls

    raise ValueError(f"Media type {value!r} not supported by any of {_readers()}")


def get_reader_for_path(path):
    """Return a reader class for file `path`.

    Raises
    ------
    ValueError
        If no reader class matches.

    See also
    --------
    BaseReader.suffixes
    """
    suffix = Path(path).suffix
    for cls in READERS:
        if cls.supports_suffix(suffix):
            return cls

    raise ValueError(f"File suffix {repr(suffix)} not supported by any of {_readers()}")


def read_sdmx(filename_or_obj, format=None, **kwargs):
    """Load a SDMX-ML or SDMX-JSON message from a file or file-like object.

    Parameters
    ----------
    filename_or_obj : str or :class:`~os.PathLike` or file
    format : 'XML' or 'JSON', optional

    Other Parameters
    ----------------
    dsd : :class:`DataStructureDefinition <.BaseDataStructureDefinition>`
        For “structure-specific” `format`=``XML`` messages only.
    """
    reader = None

    try:
        path = Path(filename_or_obj)

        # Open the file
        obj = open(path, "rb")
    except TypeError:
        # Not path-like → opened file
        path = None
        obj = filename_or_obj

    if path:
        try:
            # Use the file extension to guess the reader
            reader = get_reader_for_path(filename_or_obj)
        except ValueError:
            pass

    if not reader:
        try:
            reader = get_reader_for_path(Path(f"dummy.{format.lower()}"))
        except (AttributeError, ValueError):
            pass

    if not reader:
        # Read a line and then return the cursor to the initial position
        pos = obj.tell()
        first_line = obj.readline().strip()
        obj.seek(pos)

        try:
            reader = detect_content_reader(first_line)
        except ValueError:
            pass

    if not reader:
        raise RuntimeError(
            f"cannot infer SDMX message format from path {repr(path)}, "
            f"format={format}, or content '{first_line[:5].decode()}..'"
        )

    return reader().read_message(obj, **kwargs)



"""SDMX 3.0 Information Model."""

from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from typing import Any, ClassVar, Dict, List, Optional, Set

from . import common
from .common import (
    Code,
    Codelist,
    Component,
    ComponentList,
    ConstrainableArtefact,
    ConstraintRole,
    ConstraintRoleType,
    IdentifiableArtefact,
    MaintainableArtefact,
    NameableArtefact,
    Organisation,
    OrganisationScheme,
)
from .internationalstring import InternationalStringDescriptor

# Classes defined directly in the current file, in the order they appear
__all__ = [
    "CodelistExtension",
    "GeoRefCode",
    "GeoGridCode",
    "GeoFeatureSetCode",
    "GeoCodelist",
    "GeographicCodelist",
    "GeoGridCodelist",
    "ValueItem",
    "ValueList",
    "MetadataProvider",
    "MetadataProviderScheme",
    "Measure",
    "MeasureDescriptor",
    "DataflowRelationship",
    "MeasureRelationship",
    "ObservationRelationship",
    "DataStructureDefinition",
    "Dataflow",
    "Observation",
    "StructureSpecificDataSet",
    "MetadataAttributeDescriptor",
    "IdentifiableObjectSelection",
    "MetadataStructureDefinition",
    "Metadataflow",
    "MetadataAttributeValue",
    "CodedMetadataAttributeValue",
    "UncodedMetadataAttributeValue",
    "OtherUncodedAttributeValue",
    "TextAttributeValue",
    "XHTMLAttributeValue",
    "TargetIdentifiableObject",
    "MetadataSet",
    "Hierarchy",
    "HierarchyAssociation",
    "SelectionValue",
    "MemberValue",
    "TimeRangeValue",
    "BeforePeriod",
    "AfterPeriod",
    "RangePeriod",
    "DataKey",
    "DataKeySet",
    "Constraint",
    "MemberSelection",
    "DataConstraint",
    "MetadataConstraint",
]

# §4.3: Codelist


@dataclass
class CodeSelection:
    mv: List["MemberValue"] = field(default_factory=list)


class ExclusiveCodeSelection(CodeSelection):
    pass


class InclusiveCodeSelection(CodeSelection):
    pass


@dataclass
class CodelistExtension:
    extends: Codelist
    prefix: Optional[str] = None
    sequence: Optional[int] = None

    selection: Optional[CodeSelection] = None


class GeoRefCode(Code):
    """SDMX 3.0 GeoRefCode (abstract class)."""


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class GeoGridCode(GeoRefCode):
    """SDMX 3.0 GridCode."""

    geo_cell: str = ""  # FIXME remove the default


@dataclass
@NameableArtefact._preserve("eq", "hash", "repr")
class GeoFeatureSetCode(GeoRefCode):
    """SDMX 3.0 GeoFeatureSetCode."""

    value: str = ""  # FIXME remove the default


#: SDMX 3.0 GeoCodelistType.
GeoCodelistType = Enum("GeoCodelistType", "geographic geogrid")


@dataclass
class GeoCodelist(Codelist[GeoRefCode]):
    """SDMX 3.0 GeoCodelist (abstract class)."""

    geo_type: ClassVar[GeoCodelistType]

    _Item = GeoRefCode


@dataclass
class GeographicCodelist(GeoCodelist):
    """SDMX 3.0 GeographicCodelist."""

    geo_type = GeoCodelistType.geographic

    _Item = GeoFeatureSetCode


@dataclass
class GeoGridCodelist(GeoCodelist):
    """SDMX 3.0 GeoGridCodelist."""

    geo_type = GeoCodelistType.geogrid

    grid_definition: str = ""  # FIXME remove the default

    _Item = GeoGridCode


# §4.4: ValueList


@dataclass
@IdentifiableArtefact._preserve("hash")
class EnumeratedItem(common.AnnotableArtefact):
    """SDMX 3.0 EnumeratedItem.

    This class is similar to :meth:`.Item`, but with a only subset of its features; i.e.
    no :attr:`IdentifiableArtefact.urn`.
    """

    #: Identifier of the item.
    id: str = common.MissingID

    #: Multi-lingual name of the object. Analogous to :attr:`NameableArtefact.name`.
    name: InternationalStringDescriptor = InternationalStringDescriptor()
    #: Multi-lingual description of the object. Analogous to
    #: :attr:`NameableArtefact.description`.
    description: InternationalStringDescriptor = InternationalStringDescriptor()

    def __iter__(self, recurse=True):
        # Mirrors Item.__iter__
        yield self


class ValueItem(EnumeratedItem):
    """SDMX 3.0 ValueItem."""


class EnumeratedList(common.MaintainableArtefact):
    pass


@dataclass
class ValueList(EnumeratedList):
    """SDMX 3.0 ValueList."""

    _Item = ValueItem

    items: List[ValueItem] = field(default_factory=list)

    def append(self, item: ValueItem) -> None:
        self.items.append(item)


# §4.7: OrganisationScheme


class MetadataProvider(Organisation):
    """An organization that produces reference metadata."""


class MetadataProviderScheme(OrganisationScheme[MetadataProvider]):
    """A maintained collection of :class:`MetadataProvider`."""

    _Item = MetadataProvider


# §12.3: Constraints


@dataclass
class SelectionValue(common.BaseSelectionValue):
    #: Date from which the DataKey is valid.
    valid_from: Optional[str] = None
    #: Date from which the DataKey is superseded.
    valid_to: Optional[str] = None


class MemberValue(common.BaseMemberValue, SelectionValue):
    """SDMX 3.0 MemberValue."""


class TimeRangeValue(SelectionValue):
    """SDMX 3.0 TimeRangeValue."""


class BeforePeriod(TimeRangeValue, common.Period):
    pass


class AfterPeriod(TimeRangeValue, common.Period):
    pass


@dataclass
class RangePeriod(TimeRangeValue):
    start: Optional[common.StartPeriod] = None
    end: Optional[common.EndPeriod] = None


@dataclass
class DataKey(common.BaseDataKey):
    #: Date from which the DataKey is valid.
    valid_from: Optional[str] = None
    #: Date from which the DataKey is superseded.
    valid_to: Optional[str] = None


@dataclass
class DataKeySet(common.BaseDataKeySet):
    #: TODO the SDMX 3.0 spec is ambiguous about this: the diagram shows "member" as an
    #: attribute of :class:`.DataKey`, but the table lists it as an attribute of
    #: DataKeySet.
    member: Any = None


@dataclass
class Constraint(common.BaseConstraint):
    """SDMX 3.0 Constraint (abstract class).

    For SDMX 2.1, see :class:`.v21.Constraint`.
    """

    role: Optional[ConstraintRole] = None

    def __post_init__(self):
        if isinstance(self.role, str):
            self.role = ConstraintRole(role=ConstraintRoleType[self.role])


@dataclass
class MemberSelection(common.BaseMemberSelection):
    """SDMX 3.0 MemberSelection."""

    #: Whether Codes should retain the prefix specified in a code list extension.
    remove_prefix: bool = False


@dataclass
@NameableArtefact._preserve("repr")
class DataConstraint(Constraint):
    #:
    content: Set[ConstrainableArtefact] = field(default_factory=set)

    data_content_keys: Optional[DataKeySet] = None
    data_content_region: Optional[common.CubeRegion] = None

    def __contains__(self, name):
        raise NotImplementedError  # pragma: no cover


class MetadataConstraint(Constraint):
    metadata_content_region: Optional[common.MetadataTargetRegion] = None

    def __contains__(self, name):
        raise NotImplementedError  # pragma: no cover


# §5.3: Data Structure Definition


@dataclass
class Measure(Component):
    """SDMX 3.0 Measure.

    This class is not present in SDMX 2.1; see instead :class:`.v21.PrimaryMeasure`.
    """

    #:
    concept_role: Optional[common.Concept] = None


class MeasureDescriptor(ComponentList[Measure]):
    """SDMX 3.0 MeasureDescriptor.

    For SDMX 2.1 see instead :class:`.v21.MeasureDescriptor`.
    """

    _Component = Measure


class DataflowRelationship(common.AttributeRelationship):
    """SDMX 3.0 DataflowRelationship.

    Indicates that the attribute is attached to the entire data set. Compare with
    :class:`.v21.NoSpecifiedRelationship`.
    """


class MeasureRelationship(common.AttributeRelationship):
    """SDMX 3.0 MeasureRelationship."""


class ObservationRelationship(common.AttributeRelationship):
    """SDMX 3.0 ObservationRelationship.

    Indicates that the attribute is attached to a particular observation. Compare with
    :class:`.v21.PrimaryMeasureRelationship.`
    """


@dataclass(repr=False)
@IdentifiableArtefact._preserve("hash")
class DataStructureDefinition(common.BaseDataStructureDefinition):
    """SDMX 3.0 DataStructureDefinition (‘DSD’)."""

    MemberValue = MemberValue
    MemberSelection = MemberSelection
    ConstraintType = DataConstraint

    #: A :class:`.MeasureDescriptor`.
    measures: MeasureDescriptor = field(default_factory=MeasureDescriptor)


@dataclass(repr=False)
@IdentifiableArtefact._preserve("hash")
class Dataflow(common.BaseDataflow):
    #:
    structure: DataStructureDefinition = field(default_factory=DataStructureDefinition)


# §5.4: Data Set


@dataclass
class Observation(common.BaseObservation):
    #:
    value_for: Optional[Measure] = None


class DataSet(common.BaseDataSet):
    """SDMX 3.0 Data Set."""


class StructureSpecificDataSet(DataSet):
    """SDMX 3.0 StructureSpecificDataSet.

    This subclass has no additional functionality compared to DataSet.
    """


# §7.3 Metadata Structure Definition


class MetadataAttributeDescriptor(common.ComponentList):
    """SDMX 3.0 MetadataAttributeDescriptor."""

    _Component = common.MetadataAttribute


class IdentifiableObjectSelection:
    """SDMX 3.0 IdentifiableObjectSelection."""


@dataclass
@MaintainableArtefact._preserve("hash")
class MetadataStructureDefinition(common.BaseMetadataStructureDefinition):
    """SDMX 3.0 MetadataStructureDefinition."""

    #: A :class:`MetadataAttributeDescriptor` that describes the attributes of the
    #: metadata structure.
    #:
    #: .. note:: The SDMX 3.0.0 IM (version 1.0 / 2021-10) does not give a name for this
    #:    association. :mod:`sdmx` uses `attributes` for consistency with
    #:    :class:`.DataStructureDefinition`.
    attributes: MetadataAttributeDescriptor = field(
        default_factory=MetadataAttributeDescriptor
    )


class Metadataflow(common.BaseMetadataflow):
    """SDMX 3.0 MetadataflowDefinition."""

    structure: MetadataStructureDefinition


# §7.4: Metadata Set


class MetadataAttributeValue:
    """SDMX 3.0 MetadataAttributeValue.

    Analogous to :class:`.v21.ReportedAttribute`.
    """

    # NB the IM specifies this is a subclass of common.AttributeValue, but the
    #    implementation in .common has both Coded- and UncodedAttributeValue, which
    #    offends mypy.

    parent: Optional["MetadataAttributeValue"] = None
    child: List["MetadataAttributeValue"] = field(default_factory=list)


class CodedMetadataAttributeValue(MetadataAttributeValue):
    """SDMX 3.0 CodedMetadataAttributeValue.

    Analogous to :class:`.v21.EnumeratedAttributeValue`.
    """

    value_of: common.Code


class UncodedMetadataAttributeValue(MetadataAttributeValue):
    """SDMX 3.0 UncodedMetadataAttributeValue."""

    pass


class OtherUncodedAttributeValue(UncodedMetadataAttributeValue):
    """SDMX 3.0 OtherUncodedAttributeValue."""

    value: str
    start_time: date


class TextAttributeValue(UncodedMetadataAttributeValue, common.BaseTextAttributeValue):
    """SDMX 3.0 TextAttributeValue."""


class XHTMLAttributeValue(
    UncodedMetadataAttributeValue, common.BaseXHTMLAttributeValue
):
    """SDMX 3.0 XHTMLAttributeValue."""


class TargetIdentifiableObject:
    """SDMX 3.0 TargetIdentifiableObject."""


@dataclass
class MetadataSet(MaintainableArtefact, common.BaseMetadataSet):
    """SDMX 3.0 MetadataSet.

    .. note:: Contrast :class:`.v21.MetadataSet`, which is a :class:`.NameableArtefact`.
    """

    # NB Would prefer to type as datetime.date, but VersionableArtefact currently uses
    #    str
    valid_from: Optional[str] = None
    # NB Would prefer to type as datetime.date, but VersionableArtefact currently uses
    #    str
    valid_to: Optional[str] = None
    set_id: Optional[str] = None

    described_by: Optional[Metadataflow] = None

    # described_by: Optional[MetadataProvisionAgreement] = None

    structured_by: Optional[MetadataAttributeDescriptor] = None

    #: Analogous to :attr:`.v21.MetadataSet.published_by`.
    provided_by: Optional[MetadataProvider] = None

    attaches_to: List[TargetIdentifiableObject] = field(default_factory=list)

    metadata: List[MetadataAttributeValue] = field(default_factory=list)


# §8: Hierarchy


@dataclass
class Hierarchy(MaintainableArtefact):
    """SDMX 3.0 Hierarchy."""

    has_formal_levels: bool = False

    #: The top :class:`Level` in the hierarchy.
    level: Optional[common.Level] = None

    #: The top-level :class:`HierarchicalCodes <HierarchicalCode>` in the hierarchy.
    codes: Dict[str, common.HierarchicalCode] = field(default_factory=dict)


@dataclass
class HierarchyAssociation(MaintainableArtefact):
    """SDMX 3.0 HierarchyAssociation."""

    #: The context within which the association is performed.
    context_object: Optional[IdentifiableArtefact] = None
    #: The IdentifiableArtefact that needs the Hierarchy.
    linked_object: Optional[IdentifiableArtefact] = None
    #: The Hierarchy that is associated.
    linked_hierarchy: Optional[Hierarchy] = None


CF = common.ClassFinder(
    __name__,
    parent_map={
        Measure: MeasureDescriptor,
        common.MetadataAttribute: MetadataAttributeDescriptor,
    },
)
get_class = CF.get_class
parent_class = CF.parent_class
__dir__ = CF.dir
__getattr__ = CF.getattr



import logging
import typing
from dataclasses import fields
from typing import Generic, Tuple, TypeVar, Union, get_args, get_origin

log = logging.getLogger(__name__)

KT = TypeVar("KT")
VT = TypeVar("VT")


class DictLike(dict, typing.MutableMapping[KT, VT]):
    """Container with features of :class:`dict`, attribute access, and validation."""

    __slots__ = ("__dict__", "_types")

    def __init__(self, *args, **kwargs):
        # (key type, value type) for items
        self._types = (object, object)

        super().__init__(*args, **kwargs)

        # Ensures attribute access to dict items
        self.__dict__ = self

    @classmethod
    def with_types(cls, key_type, value_type):
        """Construct a new DictLike with the given `key_type` and `value_type`."""
        result = cls()
        result._types = (key_type, value_type)
        return result

    def __getitem__(self, key: Union[KT, int]) -> VT:
        """:meth:`dict.__getitem__` with integer access."""
        try:
            return super().__getitem__(key)
        except KeyError:
            if isinstance(key, int):
                # int() index access
                return list(self.values())[key]
            else:
                raise

    def __setitem__(self, key: KT, value: VT) -> None:
        """:meth:`dict.__setitem` with validation."""
        super().__setitem__(*self._validate_entry((key, value)))

    def __copy__(self):
        # Construct explicitly to avoid returning the parent class, dict()
        return DictLike(**self)

    def copy(self):
        """Return a copy of the DictLike."""
        return self.__copy__()

    def update(self, other):
        """Update the DictLike with elements from `other`, validating entries."""
        try:
            it = other.items()
        except AttributeError:
            it = iter(other)
        super().update(map(self._validate_entry, it))

    # Satisfy dataclass(), which otherwise complains when InternationalStringDescriptor
    # is used
    @classmethod
    def __hash__(cls):
        pass

    def _validate_entry(self, kv: Tuple):
        """Validate one `key`/`value` pair."""
        key, value = kv
        try:
            kt, vt = self._types
        except AttributeError:
            pass
        else:
            if not isinstance(key, kt):
                raise TypeError(
                    f"Expected key type {kt.__name__}; got {type(key).__name__}"
                )
            elif not isinstance(value, vt):
                raise TypeError(
                    f"Expected value type {vt.__name__}; got {type(value).__name__}"
                )

        return key, value

    def compare(self, other, strict=True):
        """Return :obj:`True` if `self` is the same as `other`.

        Two DictLike instances are identical if they contain the same set of keys, and
        corresponding values compare equal.

        Parameters
        ----------
        strict : bool, optional
            Passed to :func:`compare` for the values.
        """
        if set(self.keys()) != set(other.keys()):
            log.info(f"Not identical: {sorted(self.keys())} / {sorted(other.keys())}")
            return False

        for key, value in self.items():
            if not value.compare(other[key], strict):
                return False

        return True


# Utility methods for DictLike
#
# These are defined in separate functions to avoid collisions with keys and the
# attribute access namespace, e.g. if the DictLike contains keys "summarize" or
# "validate".


def summarize_dictlike(dl, maxwidth=72):
    """Return a string summary of the DictLike contents."""
    value_cls = dl[0].__class__.__name__
    count = len(dl)
    keys = " ".join(dl.keys())
    result = f"{value_cls} ({count}): {keys}"

    if len(result) > maxwidth:
        # Truncate the list of keys
        result = result[: maxwidth - 3] + "..."

    return result


class DictLikeDescriptor(Generic[KT, VT]):
    """Descriptor for :class:`DictLike` attributes on dataclasses."""

    def __set_name__(self, owner, name):
        self._name = "_" + name
        self._field = None
        self._types = (object, object)

    def _get_field_types(self, obj):
        """Record the types of the described field."""
        if self._field:
            return  # Already done

        # Identify the field on `obj` that matches self._name
        self._field = next(filter(lambda f: f.name == self._name[1:], fields(obj)))
        # The type is DictLike[KeyType, ValueType]; retrieve those arguments
        kt, vt = get_args(self._field.type)
        # Store. If ValueType is a generic, e.g. List[int], store only List.
        self._types = (kt, get_origin(vt) or vt)

    def __get__(self, obj, type) -> DictLike[KT, VT]:
        if obj is None:
            return None  # type: ignore [return-value]

        try:
            return obj.__dict__[self._name]
        except KeyError:
            # Construct new DictLike with specified types
            default = DictLike.with_types(*self._types)
            return obj.__dict__.setdefault(self._name, default)

    def __set__(self, obj, value):
        self._get_field_types(obj)

        if not isinstance(value, DictLike):
            # Construct new DictLike with specified types
            _value = DictLike.with_types(*self._types)
            # Update with validation
            _value.update(value or {})
            value = _value

        setattr(obj, self._name, value)



import pytest

from sdmx.reader.csv import Reader


class TestReader:
    @pytest.mark.parametrize(
        "mt, expected",
        [
            ("foo", False),
            ("application/vnd.sdmx.data+csv; version=1.0.0", True),
            ("application/vnd.sdmx.metadata+csv; version=2.0.0", True),
        ],
    )
    def test_handles_media_type(self, mt, expected) -> None:
        assert expected is Reader.handles_media_type(mt)

    @pytest.mark.parametrize("value, expected", [(".csv", True), (".xlsx", False)])
    def test_supports_suffix(self, value, expected) -> None:
        assert expected is Reader.supports_suffix(value)
